import numpy as np
import numpy.typing as npt
from shapely import Point, creation
from navsim.planning.simulation.planner.pdm_planner.observation.pdm_occupancy_map import (
    PDMDrivableMap,
)
from navsim.planning.simulation.planner.pdm_planner.utils.pdm_path import (
    PDMPath,
)
from navsim.planning.simulation.planner.pdm_planner.utils.pdm_enums import (
    StateIndex,
)
import math
from scipy.signal import savgol_filter
def _extract_ego_acceleration(
    states: npt.NDArray[np.float64],
    acceleration_coordinate: str,
    decimals: int = 8,
    poly_order: int = 2,
    window_length: int = 8,
) -> npt.NDArray[np.float32]:
    """
    Extract acceleration of ego pose in simulation history over batch-dim
    :param states: array representation of ego state values
    :param acceleration_coordinate: string of axis to extract
    :param decimals: decimal precision, defaults to 8
    :param poly_order: polynomial order, defaults to 2
    :param window_length: window size for extraction, defaults to 8
    :raises ValueError: when coordinate not available
    :return: array containing acceleration values
    """

    n_batch, n_time, n_states = states.shape
    if acceleration_coordinate == "x":
        acceleration: npt.NDArray[np.float64] = states[..., StateIndex.ACCELERATION_X]

    elif acceleration_coordinate == "y":
        acceleration: npt.NDArray[np.float64] = states[..., StateIndex.ACCELERATION_Y]

    elif acceleration_coordinate == "magnitude":
        acceleration: npt.NDArray[np.float64] = np.hypot(
            states[..., StateIndex.ACCELERATION_X],
            states[..., StateIndex.ACCELERATION_Y],
        )
    else:
        raise ValueError(
            f"acceleration_coordinate option: {acceleration_coordinate} not available. "
            f"Available options are: x, y or magnitude"
        )

    acceleration = savgol_filter(
        acceleration,
        polyorder=poly_order,
        window_length=min(window_length, n_time),
        axis=-1,
    )
    acceleration = np.round(acceleration, decimals=decimals)
    return acceleration
def _approximate_derivatives(
    y: npt.NDArray[np.float32],
    x: npt.NDArray[np.float32],
    window_length: int = 5,
    poly_order: int = 2,
    deriv_order: int = 1,
    axis: int = -1,
) -> npt.NDArray[np.float32]:
    """
    Given two equal-length sequences y and x, compute an approximation to the n-th
    derivative of some function interpolating the (x, y) data points, and return its
    values at the x's.  We assume the x's are increasing and equally-spaced.
    :param y: The dependent variable (say of length n)
    :param x: The independent variable (must have the same length n).  Must be strictly
        increasing and equally-spaced.
    :param window_length: The order (default 5) of the Savitsky-Golay filter used.
        (Ignored if the x's are not equally-spaced.)  Must be odd and at least 3
    :param poly_order: The degree (default 2) of the filter polynomial used.  Must
        be less than the window_length
    :param deriv_order: The order of derivative to compute (default 1)
    :param axis: The axis of the array x along which the filter is to be applied. Default is -1.
    :return Derivatives.
    """
    window_length = min(window_length, len(x))

    if not (poly_order < window_length):
        raise ValueError(f"{poly_order} < {window_length} does not hold!")

    dx = np.diff(x, axis=-1)
    if not (dx > 0).all():
        raise RuntimeError("dx is not monotonically increasing!")

    dx = dx.mean()
    derivative: npt.NDArray[np.float32] = savgol_filter(
        y,
        polyorder=poly_order,
        window_length=window_length,
        deriv=deriv_order,
        delta=dx,
        axis=axis,
    )
    return derivative

def _extract_ego_jerk(
    states: npt.NDArray[np.float64],
    acceleration_coordinate: str,
    time_steps_s: npt.NDArray[np.float64],
    decimals: int = 8,
    deriv_order: int = 1,
    poly_order: int = 2,
    window_length: int = 15,
) -> npt.NDArray[np.float32]:
    """
    Extract jerk of ego pose in simulation history over batch-dim
    :param states: array representation of ego state values
    :param acceleration_coordinate: string of axis to extract
    :param time_steps_s: time steps [s] of time dim
    :param decimals: decimal precision, defaults to 8
    :param deriv_order: order of derivative, defaults to 1
    :param poly_order: polynomial order, defaults to 2
    :param window_length: window size for extraction, defaults to 15
    :return: array containing jerk values
    """
    n_batch, n_time, n_states = states.shape
    ego_acceleration = _extract_ego_acceleration(
        states, acceleration_coordinate=acceleration_coordinate
    )
    jerk = _approximate_derivatives(
        ego_acceleration,
        time_steps_s,
        deriv_order=deriv_order,
        poly_order=poly_order,
        window_length=min(window_length, n_time),
    )
    jerk = np.round(jerk, decimals=decimals)
    return jerk


def _extract_ego_yaw_rate(
    states: npt.NDArray[np.float64],
    time_steps_s: npt.NDArray[np.float64],
    deriv_order: int = 1,
    poly_order: int = 2,
    decimals: int = 8,
    window_length: int = 15,
) -> npt.NDArray[np.float32]:
    """
    Extract yaw-rate of simulation history over batch-dim
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :param deriv_order: order of derivative, defaults to 1
    :param poly_order: polynomial order, defaults to 2
    :param decimals:  decimal precision, defaults to 8
    :param window_length: window size for extraction, defaults to 15
    :return: array containing ego's yaw rate
    """
    ego_headings = states[..., StateIndex.HEADING]
    ego_yaw_rate = _approximate_derivatives(
        _phase_unwrap(ego_headings),
        time_steps_s,
        deriv_order=deriv_order,
        poly_order=poly_order,
    )  # convert to seconds
    ego_yaw_rate = np.round(ego_yaw_rate, decimals=decimals)
    return ego_yaw_rate

def _phase_unwrap(headings: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    """
    Returns an array of heading angles equal mod 2 pi to the input heading angles,
    and such that the difference between successive output angles is less than or
    equal to pi radians in absolute value
    :param headings: An array of headings (radians)
    :return The phase-unwrapped equivalent headings.
    """
    # There are some jumps in the heading (e.g. from -np.pi to +np.pi) which causes approximation of yaw to be very large.
    # We want unwrapped[j] = headings[j] - 2*pi*adjustments[j] for some integer-valued adjustments making the absolute value of
    # unwrapped[j+1] - unwrapped[j] at most pi:
    # -pi <= headings[j+1] - headings[j] - 2*pi*(adjustments[j+1] - adjustments[j]) <= pi
    # -1/2 <= (headings[j+1] - headings[j])/(2*pi) - (adjustments[j+1] - adjustments[j]) <= 1/2
    # So adjustments[j+1] - adjustments[j] = round((headings[j+1] - headings[j]) / (2*pi)).
    two_pi = 2.0 * np.pi
    adjustments = np.zeros_like(headings)
    adjustments[..., 1:] = np.cumsum(
        np.round(np.diff(headings, axis=-1) / two_pi), axis=-1
    )
    unwrapped = headings - two_pi * adjustments
    return unwrapped
def _compute_lon_acceleration(
    states: npt.NDArray[np.float64], time_steps_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Compute longitudinal acceleration over batch-dim of simulated proposals
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :return: longitudinal acceleration within bound
    """
    n_batch, n_time, n_states = states.shape
    lon_acceleration = _extract_ego_acceleration(
        states, acceleration_coordinate="x", window_length=n_time
    )
    return lon_acceleration


def _compute_lat_acceleration(
    states: npt.NDArray[np.float64], time_steps_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Compute lateral acceleration over batch-dim of simulated proposals
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :return: lateral acceleration within bound
    """
    n_batch, n_time, n_states = states.shape
    lat_acceleration = _extract_ego_acceleration(
        states, acceleration_coordinate="y", window_length=n_time
    )
    return lat_acceleration
def _compute_acceleration(
    states: npt.NDArray[np.float64], time_steps_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Compute lateral acceleration over batch-dim of simulated proposals
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :return: lateral acceleration within bound
    """
    n_batch, n_time, n_states = states.shape
    lat_acceleration = _extract_ego_acceleration(
        states, acceleration_coordinate="magnitude", window_length=n_time
    )
    return lat_acceleration


def _compute_jerk_metric(
    states: npt.NDArray[np.float64], time_steps_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Compute absolute jerk over batch-dim of simulated proposals
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :return: absolute jerk within bound
    """
    n_batch, n_time, n_states = states.shape
    jerk_metric = _extract_ego_jerk(
        states,
        acceleration_coordinate="magnitude",
        time_steps_s=time_steps_s,
        window_length=n_time,
    )
    return jerk_metric


def _compute_lon_jerk_metric(
    states: npt.NDArray[np.float64], time_steps_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Compute longitudinal jerk over batch-dim of simulated proposals
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :return: longitudinal jerk within bound
    """
    n_batch, n_time, n_states = states.shape
    lon_jerk_metric = _extract_ego_jerk(
        states,
        acceleration_coordinate="x",
        time_steps_s=time_steps_s,
        window_length=n_time,
    )
    return lon_jerk_metric


def _compute_yaw_accel(
    states: npt.NDArray[np.float64], time_steps_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Compute acceleration of yaw-angle over batch-dim of simulated proposals
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :return: acceleration of yaw-angle within bound
    """
    n_batch, n_time, n_states = states.shape
    yaw_accel_metric = _extract_ego_yaw_rate(
        states, time_steps_s, deriv_order=2, poly_order=3, window_length=n_time
    )
    return yaw_accel_metric


def _compute_yaw_rate(
    states: npt.NDArray[np.float64], time_steps_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Compute velocity of yaw-angle over batch-dim of simulated proposals
    :param states: array representation of ego state values
    :param time_steps_s: time steps [s] of time dim
    :return: velocity of yaw-angle within bound
    """
    n_batch, n_time, n_states = states.shape
    yaw_rate_metric = _extract_ego_yaw_rate(states, time_steps_s, window_length=n_time)
    return yaw_rate_metric


def ego_is_comfortable(
    states: npt.NDArray[np.float64], time_point_s: npt.NDArray[np.float64]
) -> npt.NDArray[np.float32]:
    """
    Accumulates all within-bound comfortability metrics
    :param states: array representation of ego state values
    :param time_point_s: time steps [s] of time dim
    :return: _description_
    """
    n_batch, n_time, n_states = states.shape
    assert n_time == len(time_point_s)
    assert n_states == StateIndex.size()

    comfort_metric_functions = [
        # _compute_lon_acceleration,
        # _compute_lat_acceleration,
        _compute_acceleration,
        _compute_jerk_metric,
        # _compute_lon_jerk_metric,
        _compute_yaw_accel,
        _compute_yaw_rate,
    ]
    results: npt.NDArray[np.bool_] = np.zeros(
        (n_batch, n_time, len(comfort_metric_functions))
    )
    for idx, metric_function in enumerate(comfort_metric_functions):
        results[:, :, idx] = metric_function(states, time_point_s)

    return results

def compute_acceleration_consistency(acc1, acc2):
    diff = acc1 - acc2
    consistency_metric = np.sqrt(np.mean(diff**2))
    return consistency_metric

def calc_extened_comfort(states,
                         pre_states,
                         num_poses,
                         interval_length,
                         num_proposals):

    time_point_s: npt.NDArray[np.float64] = (
            np.arange(0, num_poses + 1).astype(np.float64)
            * interval_length
    )
    is_comfortable_result = ego_is_comfortable(states, time_point_s)
    is_comfortable_result_pre = ego_is_comfortable(pre_states, time_point_s)

    acc = np.zeros(num_proposals, dtype=np.float64)
    acc_lat = np.zeros(num_proposals, dtype=np.float64)
    jerk = np.zeros(num_proposals, dtype=np.float64)
    yaw_accel = np.zeros(num_proposals, dtype=np.float64)
    yaw_rate = np.zeros(num_proposals, dtype=np.float64)

    assert is_comfortable_result_pre.shape[0] == 1
    for proposal_idx in range(num_proposals):
        metric_acc_lon, metric_acc_lat, metric_jerk, metric_yaw_accel, metric_yaw_rate = 0, 0, 0, 0, 0
        for idx in range(4):
            metric_now = is_comfortable_result[proposal_idx, np.arange(35), idx]
            metric_pre = is_comfortable_result_pre[0, np.arange(5, 40), idx]
            # 0.3
            if idx == 0:
                acc[proposal_idx] = compute_acceleration_consistency(metric_now, metric_pre)
            # 0.7
            if idx == 1:
                jerk[proposal_idx] = compute_acceleration_consistency(metric_now, metric_pre)
            # 0.05
            if idx == 2:
                yaw_accel[proposal_idx] = compute_acceleration_consistency(metric_now, metric_pre)
            # 0.05
            if idx == 3:
                yaw_rate[proposal_idx] = compute_acceleration_consistency(metric_now, metric_pre)

        # for time_idx in range(35):
        #     for idx in range(5):
        #         metric_now = is_comfortable_result[proposal_idx, time_idx, idx]
        #         metric_pre = is_comfortable_result_pre[0, time_idx + 5, idx]
        #         if idx == 0:
        #             metric_acc_lon += compute_acceleration_consistency(metric_now, metric_pre)
        #         if idx == 1:
        #             metric_acc_lat += compute_acceleration_consistency(metric_now, metric_pre)
        #         if idx == 2:
        #             metric_jerk += compute_acceleration_consistency(metric_now, metric_pre)
        #         if idx == 3:
        #             metric_yaw_accel += compute_acceleration_consistency(metric_now, metric_pre)
        #         if idx == 4:
        #             metric_yaw_rate += compute_acceleration_consistency(metric_now, metric_pre)
        # acc_lon[proposal_idx] = metric_acc_lon / 35.0
        # acc_lat[proposal_idx] = metric_acc_lat / 35.0
        # jerk[proposal_idx] = metric_jerk / 35.0
        # yaw_accel[proposal_idx] = metric_yaw_accel / 35.0
        # yaw_rate[proposal_idx] = metric_yaw_rate / 35.0

    top_50_1 = np.sort(acc)[:50]
    top_50_2 = np.sort(jerk)[:50]
    top_50_3 = np.sort(yaw_accel)[:50]
    top_50_4 = np.sort(yaw_rate)[:50]
    # top_50_5 = np.sort(yaw_rate)[-50:]
    # for i in range(10):
    #     print(np.mean(top_50_1), np.mean(top_50_2), np.mean(top_50_3), np.mean(top_50_4))

    final_acc = np.ones_like(acc)
    smaller_than_mask = acc <= 0.7
    bigger_than_dis_mask = np.logical_not(smaller_than_mask)
    final_acc[bigger_than_dis_mask] = 0.7 / acc[bigger_than_dis_mask]
    # final_acc[bigger_than_dis_mask] = 0.0
    final_acc[smaller_than_mask] = 1.0

    final_jerk = np.ones_like(jerk)
    smaller_than_mask = jerk <= 0.5
    bigger_than_dis_mask = np.logical_not(smaller_than_mask)
    final_jerk[bigger_than_dis_mask] = 0.5 / jerk[bigger_than_dis_mask]
    # final_jerk[bigger_than_dis_mask] = 0.0
    final_jerk[smaller_than_mask] = 1.0

    final_yaw_accel = np.ones_like(yaw_accel)
    smaller_than_mask = yaw_accel <= 0.1
    bigger_than_dis_mask = np.logical_not(smaller_than_mask)
    final_yaw_accel[bigger_than_dis_mask] = 0.1 / yaw_accel[bigger_than_dis_mask]
    # final_yaw_accel[bigger_than_dis_mask] = 0.0
    final_yaw_accel[smaller_than_mask] = 1.0

    final_yaw_rate = np.ones_like(yaw_rate)
    smaller_than_mask = yaw_rate <= 0.1
    bigger_than_dis_mask = np.logical_not(smaller_than_mask)
    final_yaw_rate[bigger_than_dis_mask] = 0.1 / yaw_rate[bigger_than_dis_mask]
    # final_yaw_rate[bigger_than_dis_mask] = 0.0
    final_yaw_rate[smaller_than_mask] = 1.0


    fina_weighed_score = final_acc * final_jerk * final_yaw_accel * final_yaw_rate
    # assert np.all(fina_weighed_score == 1.0) == False

    return fina_weighed_score
    # self._weighted_metrics[WeightedMetricIndex.COMFORTABLE] = np.all(is_comfortable, axis=-1)

