import logging
import lzma
import os
import pickle
import traceback
import uuid
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Union, Tuple

import hydra
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.distributed as dist
from hydra.utils import instantiate
from nuplan.common.actor_state.state_representation import StateSE2
from nuplan.common.actor_state.vehicle_parameters import get_pacifica_parameters
from nuplan.planning.script.builders.logging_builder import build_logger
from nuplan.planning.utils.multithreading.worker_utils import worker_map
from omegaconf import DictConfig
from torch.utils.data import DataLoader

from navsim.agents.abstract_agent import AbstractAgent
from navsim.agents.hydra_plantf.hydra_plantf_features import HydraPlantfFeatureBuilder
from navsim.common.dataloader import MetricCacheLoader
from navsim.common.dataloader import SceneLoader, SceneFilter
from navsim.evaluate.pdm_score import pdm_score
from navsim.planning.metric_caching.metric_cache import MetricCache
from navsim.planning.script.builders.worker_pool_builder import build_worker
from navsim.planning.script.lctgen.convert_lctgen_agents import build_lctgen_obs
from navsim.planning.simulation.planner.pdm_planner.scoring.pdm_scorer import PDMScorer
from navsim.planning.simulation.planner.pdm_planner.simulation.pdm_simulator import (
    PDMSimulator
)
from navsim.planning.training.agent_lightning_module import AgentLightningModule
from navsim.planning.training.dataset import Dataset

"""
ckpt -> pkl + valid score

"""

logger = logging.getLogger(__name__)

CONFIG_PATH = "../config/pdm_scoring"
CONFIG_NAME = "run_pdm_score_ddp"

vehicle_params = get_pacifica_parameters()
# frame 0: log
# frame 0-18: lctgen history
# frame 19 is the key frame used for evaluation
key_frame_idx = 9
ego_idx = 0
log_frame_idx = 3

def get_ego_state_at_frame(ori_ego_pose, key_frame_idx, traj, heading):
    """
    ori_ego_pose: global ego pose at frame 0
    """
    return StateSE2(
        ori_ego_pose[0] + traj[key_frame_idx, ego_idx, 0].item(),
        ori_ego_pose[1] + traj[key_frame_idx, ego_idx, 1].item(),
        ori_ego_pose[2] + heading[key_frame_idx, ego_idx].item(),
    )


@hydra.main(config_path=CONFIG_PATH, config_name=CONFIG_NAME)
def main(cfg: DictConfig) -> None:
    build_logger(cfg)
    # gpu inference
    agent: AbstractAgent = instantiate(cfg.agent)
    agent.initialize()
    scene_filter = instantiate(cfg.scene_filter)
    scene_loader = SceneLoader(
        sensor_blobs_path=Path(cfg.sensor_blobs_path),
        data_path=Path(cfg.navsim_log_path),
        scene_filter=scene_filter,
        sensor_config=agent.get_sensor_config(),
    )
    metric_cache_loader = MetricCacheLoader(Path(cfg.metric_cache_path))

    tokens_to_evaluate = list(set(scene_loader.tokens) & set(metric_cache_loader.tokens))
    num_missing_metric_cache_tokens = len(set(scene_loader.tokens) - set(metric_cache_loader.tokens))
    num_unused_metric_cache_tokens = len(set(metric_cache_loader.tokens) - set(scene_loader.tokens))
    if num_missing_metric_cache_tokens > 0:
        logger.warning(f"Missing metric cache for {num_missing_metric_cache_tokens} tokens. Skipping these tokens.")
    if num_unused_metric_cache_tokens > 0:
        logger.warning(f"Unused metric cache for {num_unused_metric_cache_tokens} tokens. Skipping these tokens.")
    logger.info("Starting pdm scoring of %s scenarios...", str(len(tokens_to_evaluate)))

    val_data = Dataset(
        scene_loader=scene_loader,
        feature_builders=agent.get_feature_builders(),
        target_builders=agent.get_target_builders(),
        cache_path=None,
        force_cache_computation=False,
        append_token_to_batch=True
    )
    all_lctgen_data = {}
    for data in val_data:
        features, _, token = data
        scene = val_data._scene_loader.get_scene_from_token(token)

        lctgen_data = torch.load(f'/mnt/f/e2e/navsim_ours/debug/plantf/lctgen/{token}.pth')
        all_lctgen_data[token] = lctgen_data
        length_width, traj, heading, vel = (
            lctgen_data['length_width'],
            lctgen_data['traj'],
            lctgen_data['heading'],
            lctgen_data['vel']
        )
        agent_cnt = length_width.shape[0] - 1
        new_states, new_categories, new_valid_mask, new_status_feature = (
            torch.zeros_like(features['agent']['states']),
            torch.zeros_like(features['agent']['categories']),
            torch.zeros_like(features['agent']['valid_mask']),
            torch.zeros_like(features['status_feature'])
        )
        # 19 is the key frame
        new_states[:agent_cnt] = torch.cat([
            traj[key_frame_idx, 1:],  # x, y
            heading[key_frame_idx, 1:].cos(),
            heading[key_frame_idx, 1:].sin(),
            length_width[1:],  # L, W
            vel[key_frame_idx, 1:, ]  # vx, vy
        ], dim=-1)
        new_valid_mask[:agent_cnt] = 1
        new_status_feature[:4] = features['status_feature'][:4]
        new_status_feature[4:6] = vel[key_frame_idx, 0]
        new_status_feature[6:8] = (vel[key_frame_idx, 0] - vel[key_frame_idx - 1, 0]) / 0.1
        features['agent']['states'] = new_states
        # all agents are vehicles
        features['agent']['categories'] = new_categories
        features['agent']['valid_mask'] = new_valid_mask
        features['status_feature'] = new_status_feature
        feature_builder: HydraPlantfFeatureBuilder = agent.get_feature_builders()[0]
        ori_ego_pose = scene.frames[log_frame_idx].ego_status.ego_pose
        new_state_se2 = get_ego_state_at_frame(
            ori_ego_pose, key_frame_idx, traj, heading
        )
        features['map'] = feature_builder._compute_map_features(
            scene.map_api, new_state_se2
        )

    val_dataloader = DataLoader(val_data, **cfg.dataloader.params, shuffle=False)
    logger.info("Num validation samples: %d", len(val_data))
    assert len(val_data) == len(tokens_to_evaluate), f'dataloader: {len(val_data)}, tokens: {len(tokens_to_evaluate)}'

    trainer = pl.Trainer(**cfg.trainer.params, callbacks=agent.get_training_callbacks())

    logger.info("Starting Training")
    predictions = trainer.predict(
        AgentLightningModule(
            agent=agent,
        ),
        val_dataloader,
        return_predictions=True
    )
    dist.barrier()
    all_predictions = [None for _ in range(dist.get_world_size())]

    if dist.is_initialized():
        dist.all_gather_object(all_predictions, predictions)
    else:
        all_predictions.append(predictions)

    if dist.get_rank() == 0:
        merged_predictions = {}
        for proc_prediction in all_predictions:
            for d in proc_prediction:
                merged_predictions.update(d)
        agent_ckpt_path = Path(cfg.agent.checkpoint_path).parent.absolute().__str__()
        ckpt_name = Path(cfg.agent.checkpoint_path).name.split('.')[0]
        pickle.dump(merged_predictions, open(f'{agent_ckpt_path}/{ckpt_name}.pkl', 'wb'))

        data_points = [
            {
                "cfg": cfg,
                "log_file": log_file,
                "tokens": tokens_list,
                "model_trajectory": merged_predictions,
                "lctgen_data": all_lctgen_data
            }
            for log_file, tokens_list in scene_loader.get_tokens_list_per_log().items()
        ]
        total_token_cnt = sum([len(t["tokens"]) for t in data_points])
        assert len(merged_predictions) == total_token_cnt, (f'merged: {len(merged_predictions)},'
                                                            f'total: {total_token_cnt}')

        worker = build_worker(cfg)
        score_rows: List[Tuple[Dict[str, Any], int, int]] = worker_map(worker, run_pdm_score, data_points)
        pdm_score_df = pd.DataFrame(score_rows)
        num_sucessful_scenarios = pdm_score_df["valid"].sum()
        num_failed_scenarios = len(pdm_score_df) - num_sucessful_scenarios
        average_row = pdm_score_df.drop(columns=["token", "valid"]).mean(skipna=True)
        average_row["token"] = "average"
        average_row["valid"] = pdm_score_df["valid"].all()
        pdm_score_df.loc[len(pdm_score_df)] = average_row

        save_path = Path(cfg.output_dir)
        timestamp = datetime.now().strftime("%Y.%m.%d.%H.%M.%S")
        pdm_score_df.to_csv(save_path / f"{timestamp}.csv")

        logger.info(f"""
            Finished running evaluation.
                Number of successful scenarios: {num_sucessful_scenarios}. 
                Number of failed scenarios: {num_failed_scenarios}.
                Final average score of valid results: {pdm_score_df['score'].mean()}.
                Results are stored in: {save_path / f"{timestamp}.csv"}.
        """)


def run_pdm_score(args: List[Dict[str, Union[List[str], DictConfig]]]) -> List[Dict[str, Any]]:
    node_id = int(os.environ.get("NODE_RANK", 0))
    thread_id = str(uuid.uuid4())
    logger.info(f"Starting worker in thread_id={thread_id}, node_id={node_id}")

    log_names = [a["log_file"] for a in args]
    tokens = [t for a in args for t in a["tokens"]]
    cfg: DictConfig = args[0]["cfg"]
    all_lctgen_data = args[0]['lctgen_data']

    simulator: PDMSimulator = instantiate(cfg.simulator)
    scorer: PDMScorer = instantiate(cfg.scorer)
    assert simulator.proposal_sampling == scorer.proposal_sampling, "Simulator and scorer proposal sampling has to be identical"

    metric_cache_loader = MetricCacheLoader(Path(cfg.metric_cache_path))
    scene_filter: SceneFilter = instantiate(cfg.scene_filter)
    scene_filter.log_names = log_names
    scene_filter.tokens = tokens
    scene_loader = SceneLoader(
        sensor_blobs_path=Path(cfg.sensor_blobs_path),
        data_path=Path(cfg.navsim_log_path),
        scene_filter=scene_filter,
    )
    model_trajectory = args[0]['model_trajectory']
    tokens_to_evaluate = list(set(scene_loader.tokens) & set(metric_cache_loader.tokens))
    pdm_results: List[Dict[str, Any]] = []
    for idx, (token) in enumerate(tokens_to_evaluate):
        logger.info(
            f"Processing scenario {idx + 1} / {len(tokens_to_evaluate)} in thread_id={thread_id}, node_id={node_id}"
        )
        score_row: Dict[str, Any] = {"token": token, "valid": True}
        try:
            # todo metric cache -> modify pdm traj / observations
            metric_cache_path = metric_cache_loader.metric_cache_paths[token]
            with lzma.open(metric_cache_path, "rb") as f:
                metric_cache: MetricCache = pickle.load(f)
            # override observation in metric cache
            scene = scene_loader.get_scene_from_token(token)
            lctgen_data = all_lctgen_data[token]

            ori_ego_pose = scene.frames[log_frame_idx].ego_status.ego_pose
            init_ego_state = StateSE2(
                ori_ego_pose[0],
                ori_ego_pose[1],
                ori_ego_pose[2]
            )
            metric_cache.observation = build_lctgen_obs(lctgen_data,
                                                        init_ego_state,
                                                        key_frame_idx,
                                                        scene.frames[log_frame_idx].timestamp)
            pdm_result = pdm_score(
                metric_cache=metric_cache,
                model_trajectory=model_trajectory[token]['trajectory'],
                future_sampling=simulator.proposal_sampling,
                simulator=simulator,
                scorer=scorer,
                use_pdm_closed=cfg.get('use_pdm_closed', False),
            )
            score_row.update(asdict(pdm_result))
        except Exception as e:
            logger.warning(f"----------- Agent failed for token {token}:")
            traceback.print_exc()
            score_row["valid"] = False

        pdm_results.append(score_row)
    return pdm_results


if __name__ == "__main__":
    main()
